@@ -1,7 +1,7 @@
 #!/usr/bin/env perl
 # vim:ts=4:sw=4:expandtab
 #
-# i3 - an improved tiling window manager
+# i3 - an improved dynamic tiling window manager
 # © 2009 Michael Stapelberg and contributors (see also: LICENSE)
 #
 # generate-command-parser.pl: script to generate parts of the command parser
@@ -147,12 +147,11 @@ for my $state (@keys) {
         $next_state ||= 'INITIAL';
         my $fmt = $cmd;
         # Replace the references to identified literals (like $workspace) with
-        # calls to parser_get_string(). Also replaces state names (like
-        # FOR_WINDOW) with their ID (useful for cfg_criteria_init(FOR_WINDOW)
-        # e.g.).
+        # calls to get_string(). Also replaces state names (like FOR_WINDOW)
+        # with their ID (useful for cfg_criteria_init(FOR_WINDOW) e.g.).
         $cmd =~ s/$_/$statenum{$_}/g for @keys;
-        $cmd =~ s/\$([a-z_]+)/parser_get_string(stack, "$1")/g;
-        $cmd =~ s/\&([a-z_]+)/parser_get_long(stack, "$1")/g;
+        $cmd =~ s/\$([a-z_]+)/get_string(stack, "$1")/g;
+        $cmd =~ s/\&([a-z_]+)/get_long(stack, "$1")/g;
         # For debugging/testing, we print the call using printf() and thus need
         # to generate a format string. The format uses %d for <number>s,
         # literal numbers or state IDs and %s for NULL, <string>s and literal
