@@ -46,6 +46,7 @@
 #include "click.h"
 #include "key_press.h"
 #include "floating.h"
+#include "gaps.h"
 #include "drag.h"
 #include "configuration.h"
 #include "handlers.h"
@@ -58,6 +59,7 @@
 #include "match.h"
 #include "xcursor.h"
 #include "resize.h"
+#include "tiling_drag.h"
 #include "sighandler.h"
 #include "move.h"
 #include "output.h"
